import { HttpModule, HttpService } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { ConfigModule } from '@nestjs/config';
import { getModelToken } from '@nestjs/mongoose';
import { Test, TestingModule } from '@nestjs/testing';
import * as bcrypt from 'bcrypt';
import { Types } from 'mongoose';
import { of } from 'rxjs';
import { personalOffersDataMock } from '../../../test/mock/data/personalOffers.mock.data';
import {
  mockStructureDocument,
  structureMockDto,
  structuresDocumentDataMock,
} from '../../../test/mock/data/structures.mock.data';
import { userDetails } from '../../../test/mock/data/users.mock.data';
import { mockParametersModel } from '../../../test/mock/services/parameters.mock.service';
import {
  mockCNFSStructures,
  mockResinStructures,
  mockSearchAdressBal,
  mockSearchAdressBan,
} from '../../../test/mock/services/structures.mock.service';
import { UsersServiceMock } from '../../../test/mock/services/user.mock.service';
import { CategoriesService } from '../../categories/services/categories.service';
import { ConfigurationService } from '../../configuration/configuration.service';
import { MailerModule } from '../../mailer/mailer.module';
import { MailerService } from '../../mailer/mailer.service';
import { ParametersService } from '../../parameters/parameters.service';
import { Parameters } from '../../parameters/schemas/parameters.schema';
import { PersonalOfferDocument } from '../../personal-offers/schemas/personal-offer.schema';
import { SearchModule } from '../../search/search.module';
import { IUser } from '../../users/interfaces/user.interface';
import { UsersService } from '../../users/services/users.service';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { StructuresSearchService } from './structures-search.service';
import { StructuresService } from './structures.service';

function hashPassword() {
  return bcrypt.hashSync(process.env.USER_PWD, process.env.SALT);
}

jest.mock('axios');

const mockStructureModel = {
  countDocuments: jest.fn(),
  create: jest.fn(),
  deleteOne: jest.fn(),
  exec: jest.fn(),
  find: jest.fn(),
  findById: jest.fn(),
  findByIdAndUpdate: jest.fn(),
  findOne: jest.fn(),
  limit: jest.fn(),
  populate: jest.fn(),
  save: jest.fn(),
  select: jest.fn().mockReturnThis(),
};

const structuresSearchServiceMock = {
  search: jest.fn().mockReturnValue([
    {
      _id: '6903ba0e2ab5775cfc01ed4d',
      structureId: '6903ba0e2ab5775cfc01ed4d',
      structureType: null,
      digitalCultureSecurity: ['2', '5', '9', '28', '34', '39', '42', '51', '52', '54', '65', '96', '97', '98'],
      parentingHelp: ['3', '22', '82', '94'],
      socialAndProfessional: ['6', '20', '66', '67', '68', '69', '124', '125', '127'],
      accessRight: ['84', '85', '86', '87', '88', '89', '93', '95'],
      baseSkills: ['260', '1', '11', '38', '48', '74', '77'],
      proceduresAccompaniment: ['cpam', 'impots', 'carsat', 'poleEmploi'],
      publics: ['toutPublic'],
      labelsQualifications: ['passNumerique', 'espacePublicNumeriqueepn'],
      accessModality: ['accesLibre', 'telephoneVisio', 'surRdv'],
      freeWorkShop: 'Non',
      createdAt: '2020-11-16T09:30:00.000Z',
      updatedAt: '2021-04-12T08:48:00.000Z',
      structureName: "L'Atelier Numérique",
      description:
        "L'Atelier Numérique est l'Espace Public Numérique des Centres Sociaux de Rillieux-la-Pape, ayant pour mission la médiation numérique pour toutes et tous.",
      lockdownActivity:
        'accesLibres, permanences numériques téléphoniques, cours et ateliers à distance, formations professionnelles.',
      contactPhone: '',
      contactMail: '',
      website: '',
      facebook: null,
      twitter: null,
      instagram: null,
      pmrAccess: true,
      exceptionalClosures: '',
      jaccompagneLesUsagersDansLeursDemarchesEnLigne: true,
      publicsAccompaniment: [],
      autresAccompagnements: '',
      nbComputers: 16,
      nbPrinters: 1,
      nbTablets: 1,
      nbNumericTerminal: 1,
      hours: {
        monday: {
          open: true,
          time: [
            {
              closing: '12:30',
              opening: '09:00',
            },
            {
              closing: '17:00',
              opening: '13:30',
            },
          ],
        },
        tuesday: {
          open: true,
          time: [
            {
              closing: '12:30',
              opening: '09:00',
            },
            {
              closing: '17:00',
              opening: '13:30',
            },
          ],
        },
        wednesday: {
          open: true,
          time: [
            {
              closing: '12:30',
              opening: '09:00',
            },
            {
              closing: '17:00',
              opening: '13:30',
            },
          ],
        },
        thursday: {
          open: true,
          time: [
            {
              closing: '12:30',
              opening: '09:00',
            },
            {
              closing: '17:00',
              opening: '13:30',
            },
          ],
        },
        friday: {
          open: true,
          time: [
            {
              closing: '12:30',
              opening: '09:00',
            },
          ],
        },
        saturday: {
          open: false,
          time: [],
        },
        sunday: {
          open: false,
          time: [],
        },
      },
      __v: 0,
      address: {
        numero: '30 bis',
        street: 'Avenue Leclerc',
        commune: 'Rillieux-la-Pape',
        postcode: '69140',
        inseeCode: '69286',
      },
      coord: [4.9036773, 45.8142196],
      accountVerified: true,
      linkedin: null,
      nbScanners: 1,
      otherDescription: null,
      personalOffers: [],
    },
  ]),
  dropIndex: jest.fn(),
  createStructureIndex: jest.fn(),
  indexStructure: jest.fn(),
};

const mockCategoriesService = {
  findOneComplete: jest.fn().mockReturnValue({
    _id: '5fbb92ef80a5c257dc0161f2',
    modules: [
      {
        id: 'computer',
        name: "Prise en main de l'ordinateur",
      },
      {
        id: 'internet',
        name: "Utilisation d'Internet",
      },
      {
        id: 'smartphone',
        name: 'Prise en main du smartphone, de la tablette',
      },
      {
        id: 'mail',
        name: 'Création et gestion de sa boîte mail',
      },
      {
        id: 'work',
        name: "Recherche d'emploi en ligne",
      },
      {
        id: 'officeAutomation',
        name: 'Bureautique et mise en page',
      },
      {
        id: 'communication',
        name: 'Outils de communication (réseaux sociaux, visio...)',
      },
      {
        id: 'data',
        name: 'Gérer ses données',
      },
      {
        id: 'scolarity',
        name: 'Scolarité (Pronote, laclasse.com...)',
      },
    ],
    name: 'Compétences numériques de base',
    theme: 'Compétences numériques',
    id: 'baseSkills',
    __v: 0,
  }),
};

const httpServiceMock = {
  get: jest.fn(),
};

describe('StructuresService', () => {
  let structuresService: StructuresService;

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [HttpModule, MailerModule, SearchModule, ConfigModule],
      providers: [
        StructuresService,
        ConfigurationService,
        ParametersService,
        {
          provide: HttpService,
          useValue: httpServiceMock,
        },
        {
          provide: StructuresSearchService,
          useValue: structuresSearchServiceMock,
        },
        {
          provide: getModelToken(Structure.name),
          useValue: mockStructureModel,
        },
        {
          provide: CategoriesService,
          useValue: mockCategoriesService,
        },
        {
          provide: UsersService,
          useClass: UsersServiceMock,
        },
        {
          provide: getModelToken(Parameters.name),
          useValue: mockParametersModel,
        },
      ],
    }).compile();

    structuresService = module.get<StructuresService>(StructuresService);
    structuresService['structuresSearchService']['index'] = 'structures-unit-test';
    structuresService['sendToBeDeletedNotification'] = jest.fn();
  });

  it('should be defined', () => {
    expect(structuresService).toBeDefined();
  });

  it('should Initiate structure index', async () => {
    await structuresService.initiateStructureIndex();
    expect(structuresSearchServiceMock.createStructureIndex).toBeCalledTimes(1);
    expect(structuresSearchServiceMock.dropIndex).toBeCalledTimes(1);
    expect(structuresSearchServiceMock.indexStructure).toBeCalledTimes(1);
  });

  describe('searchForStructures', () => {
    jest.setTimeout(30000);
    mockStructureModel.find.mockReturnThis();
    mockStructureModel.limit.mockReturnThis();
    mockStructureModel.populate.mockReturnThis();
    mockStructureModel.exec.mockResolvedValue([
      {
        _id: '6903ba0e2ab5775cfc01ed4d',
        structureId: '6903ba0e2ab5775cfc01ed4d',
        structureType: null,
        digitalCultureSecurity: ['2', '5', '9', '28', '34', '39', '42', '51', '52', '54', '65', '96', '97', '98'],
        parentingHelp: ['3', '22', '82', '94'],
        socialAndProfessional: ['6', '20', '66', '67', '68', '69', '124', '125', '127'],
        accessRight: ['84', '85', '86', '87', '88', '89', '93', '95'],
        baseSkills: ['260', '1', '11', '38', '48', '74', '77'],
        proceduresAccompaniment: ['cpam', 'impots', 'carsat', 'poleEmploi'],
        publics: ['toutPublic'],
        labelsQualifications: ['passNumerique', 'espacePublicNumeriqueepn'],
        accessModality: ['accesLibre', 'telephoneVisio', 'surRdv'],
        freeWorkShop: 'Non',
        createdAt: '2020-11-16T09:30:00.000Z',
        updatedAt: '2021-04-12T08:48:00.000Z',
        structureName: "L'Atelier Numérique",
        description:
          "L'Atelier Numérique est l'Espace Public Numérique des Centres Sociaux de Rillieux-la-Pape, ayant pour mission la médiation numérique pour toutes et tous.",
        lockdownActivity:
          'accesLibres, permanences numériques téléphoniques, cours et ateliers à distance, formations professionnelles.',
        contactPhone: '',
        contactMail: '',
        website: '',
        facebook: null,
        twitter: null,
        instagram: null,
        pmrAccess: true,
        exceptionalClosures: '',
        publicsAccompaniment: [],
        autresAccompagnements: '',
        nbComputers: 16,
        nbPrinters: 1,
        nbTablets: 1,
        nbNumericTerminal: 1,
        hours: {
          monday: {
            open: true,
            time: [
              {
                closing: '12:30',
                opening: '09:00',
              },
              {
                closing: '17:00',
                opening: '13:30',
              },
            ],
          },
        },
        __v: 0,
        address: {
          numero: '30 bis',
          street: 'Avenue Leclerc',
          commune: 'Rillieux-la-Pape',
          postcode: '69140',
          inseeCode: '69286',
        },
        coord: [4.9036773, 45.8142196],
        accountVerified: true,
        linkedin: null,
        nbScanners: 1,
        otherDescription: null,
        personalOffers: [],
      },
    ]);

    it('should find 1 structure', async () => {
      const res = await structuresService.searchForStructures('a', [{ nbPrinters: '1' }]);
      expect(res.length).toBe(1);
    });
    it('should find 1 structure', async () => {
      const res = await structuresService.searchForStructures('a', [{ nbPrinters: '1' }, { '': 'baseSkills' }]);
      expect(res.length).toBe(1);
    });
    it('should find 1 structure', async () => {
      const res = await structuresService.searchForStructures('a', [{ '': 'baseSkills' }]);
      expect(res.length).toBe(1);
    });
    it('should find 1 structure', async () => {
      const res = await structuresService.searchForStructures('a');
      expect(res.length).toBe(1);
    });
  });

  describe('create', () => {
    it('should return invalid profile', async () => {
      try {
        await structuresService.create('test@test.com', structureMockDto);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Invalid profile');
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });
  });

  it('should set structure to be deleted', async () => {
    const res = await structuresService.setToBeDeleted(userDetails[0], structuresDocumentDataMock[0]);
    expect(res.toBeDeletedAt).toBeTruthy();
  });

  it('should cancel structure delete', async () => {
    const res = await structuresService.cancelDelete(userDetails[0], structuresDocumentDataMock[0]);
    expect(res.toBeDeletedAt).toBeNull();
  });

  it('should search structure', () => {
    const filters = [{ nbPrinters: '1' }];
    let res = structuresService.search('', filters);
    expect(res).toBeTruthy();
    res = structuresService.search(null, filters);
    expect(res).toBeTruthy();
    res = structuresService.search(null);
    expect(res).toBeTruthy();
  });

  describe('findAll', () => {
    it('should find all structures', () => {
      expect(structuresService.findAll()).toBeTruthy();
    });

    it('should find all deleted structures', () => {
      expect(structuresService.findAll(true)).toBeTruthy();
    });

    it('should find all unclaimed structures', () => {
      expect(structuresService.findAllUnclaimed()).toBeTruthy();
    });

    it('should find all formatted structures', () => {
      mockStructureModel.exec.mockResolvedValue([]);
      expect(structuresService.findConsentingStructures(null)).toBeTruthy();
    });
  });

  it('should populate ES', () => {
    expect(structuresService.populateES()).toBeTruthy();
  });

  describe('getAllDataConsentPendingStructures', () => {
    it('should get one empty consent', async () => {
      const user = {
        _id: '123',
        validationToken:
          'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
        emailVerified: false,
        email: 'jacques.dupont@mii.com',
        password: hashPassword(),
        role: 0,
        newEmail: '',
        changeEmailToken: '',
        resetPasswordToken: null,
        structuresLink: [
          new Types.ObjectId('620e510b7d8f26712c386be7'),
          new Types.ObjectId('620e51247d8f26712c386be9'),
          new Types.ObjectId('620e5236f25755550cb86dfd'),
        ],
        pendingStructuresLink: [],
        structureOutdatedMailSent: [],
        personalOffers: [],
        name: 'Jacques',
        surname: 'Dupont',
        phone: '06 06 06 06 06',
        createdAt: new Date('2022-05-25T09:48:28.824Z'),
      } as IUser;
      jest
        .spyOn(structuresService, 'findOne')
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('61e9260b2ac971550065e261'),
          coord: [4.844309, 45.865288],
          createdAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          updatedAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          __v: 0,
        } as StructureDocument)
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('61e9260b2ac971550065e261'),
          coord: [4.844309, 45.865288],
          createdAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          updatedAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          dataShareConsentDate: null,
        } as StructureDocument)
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('61e9260b2ac971550065e261'),
          coord: [4.844309, 45.865288],
          createdAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          updatedAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          dataShareConsentDate: new Date(),
        } as StructureDocument);
      const res = await structuresService.getAllDataConsentPendingStructures(user);
      expect(res.length).toBe(1);
    });
    it('should get no consent', async () => {
      const user = {
        _id: '123',
        validationToken:
          'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
        emailVerified: false,
        email: 'jacques.dupont@mii.com',
        password: hashPassword(),
        role: 0,
        newEmail: '',
        changeEmailToken: '',
        resetPasswordToken: null,
        structuresLink: [
          new Types.ObjectId('620e510b7d8f26712c386be7'),
          new Types.ObjectId('620e51247d8f26712c386be9'),
          new Types.ObjectId('620e5236f25755550cb86dfd'),
        ],
        pendingStructuresLink: [],
        structureOutdatedMailSent: [],
        personalOffers: [],
        name: 'Jacques',
        surname: 'Dupont',
        phone: '06 06 06 06 06',
        createdAt: new Date('2022-05-25T09:48:28.824Z'),
      } as IUser;
      jest
        .spyOn(structuresService, 'findOne')
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('61e9260b2ac971550065e261'),
          coord: [4.844309, 45.865288],
          createdAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          updatedAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          dataShareConsentDate: null,
          __v: 0,
        } as StructureDocument)
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('61e9260b2ac971550065e261'),
          coord: [4.844309, 45.865288],
          createdAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          updatedAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          dataShareConsentDate: null,
        } as StructureDocument)
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('61e9260b2ac971550065e261'),
          coord: [4.844309, 45.865288],
          createdAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          updatedAt: new Date('Thu Jan 20 2022 10:06:19 GMT+0100 (heure normale d’Europe centrale)'),
          dataShareConsentDate: new Date(),
        } as StructureDocument);
      const res = await structuresService.getAllDataConsentPendingStructures(user);
      expect(res.length).toBe(0);
    });
  });

  it('should report structure Error', () => {
    jest.spyOn(structuresService, 'findOne').mockResolvedValue(mockStructureDocument);
    jest.spyOn(MailerService.prototype, 'send').mockImplementation(jest.fn());
    let res = structuresService.reportStructureError('6093ba0e2ab5775cfc01ed3e', '');
    expect(res).toBeTruthy();
    res = structuresService.reportStructureError(null, '');
    expect(res).toBeTruthy();
  });

  describe('addPersonalOffer', () => {
    const personalOfferDocumentMock: PersonalOfferDocument = personalOffersDataMock[0] as PersonalOfferDocument;
    it('should add personal offer to the structure', async () => {
      jest.spyOn(structuresService, 'findOne').mockResolvedValue(structuresDocumentDataMock[0]);
      const expectedResult = { ...structuresDocumentDataMock[0], personalOffers: [personalOfferDocumentMock] };
      expect(await structuresService.addPersonalOffer('6093ba0e2ab5775cfc01ed3e', personalOfferDocumentMock)).toEqual(
        expectedResult
      );
    });
    it('should return exception if structure is not found for given id', async () => {
      jest.spyOn(structuresService, 'findOne').mockResolvedValue(null);
      try {
        await structuresService.addPersonalOffer('abcd', personalOfferDocumentMock);
        // Fail test if above expression doesn't throw anything.
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Structure not found for the personal offer attachment');
        expect(e.status).toBe(400);
      }
    });
    it('should return exception if personal offer already exists in the structure', async () => {
      jest.spyOn(structuresService, 'findOne').mockResolvedValue(structuresDocumentDataMock[1]);
      try {
        await structuresService.addPersonalOffer('6093ba0e2ab5775cfc01ed3e', personalOfferDocumentMock);
        // Fail test if above expression doesn't throw anything.
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Personal offer already exist in the structure');
        expect(e.status).toBe(400);
      }
    });
  });

  describe('CNFS Structures', () => {
    beforeEach(() => {
      httpServiceMock.get.mockImplementationOnce(() => of({ data: mockCNFSStructures }));
    });

    it('getCNFSStructuress(): should return 2 structures from the mock containing 4 structures', async () => {
      expect((await structuresService.getCNFSStructures()).length).toBe(2);
    });
    it('bindOneCNFSStructure(): should update ONE structure', async () => {
      const mockSetCNFSid = jest.fn().mockResolvedValue(mockResinStructures[0]);
      jest.spyOn(structuresService, 'findOne').mockResolvedValue(mockResinStructures[0] as StructureDocument);
      jest.spyOn(structuresService, 'setCNFSid').mockImplementation(mockSetCNFSid);
      mockStructureModel.findByIdAndUpdate.mockReturnThis();
      mockStructureModel.exec.mockResolvedValueOnce(mockResinStructures[0]);
      expect(await structuresService.bindOneCNFSStructure('mockId')).toBe('A match has been found, updating CNFSid');
    });
    it('bindOneCNFSStructure(): should update ZERO structure, no match', async () => {
      jest.spyOn(structuresService, 'findOne').mockReturnThis();
      jest.spyOn(structuresService, 'update').mockResolvedValue(mockResinStructures[2]);
      expect(await structuresService.bindOneCNFSStructure('mockId')).toBe('No match found');
    });
    it('bindCNFSids(): should update 2 structures (matching phone and mail)', async () => {
      const mockSetCNFSid = jest.fn().mockResolvedValue(mockResinStructures[0]);
      jest.spyOn(structuresService, 'findAll').mockResolvedValue(mockResinStructures as StructureDocument[]);
      jest.spyOn(structuresService, 'setCNFSid').mockImplementation(mockSetCNFSid);
      expect(await structuresService.bindCNFSids()).toBe(`2 structures affected`);
      expect(mockSetCNFSid).toHaveBeenCalledTimes(2);
    });
  });

  describe('restoreStructure', () => {
    it('should restore a structure', async () => {
      mockStructureModel.findByIdAndUpdate.mockReturnThis();
      mockStructureModel.exec.mockImplementation(() => null);
      await structuresService.restoreStructure('someId');
      expect(mockStructureModel.findByIdAndUpdate).toBeCalled();
    });
  });

  describe('searchAddress', () => {
    beforeEach(() => {
      httpServiceMock.get.mockClear();
      httpServiceMock.get
        .mockImplementationOnce(() => of({ data: mockSearchAdressBan }))
        .mockImplementationOnce(() => of({ data: mockSearchAdressBal }));
    });

    it('should return only one result for duplicate addresses and addresses with diacritical marks', async () => {
      const data = { searchQuery: 'rue édison' };
      const expectedResult = {
        features: [
          {
            geometry: { coordinates: [4.993358, 45.7753763], type: 'Point' },
            type: 'Feature',
            properties: {
              osm_id: 13529,
              osm_type: 'N',
              country: 'France',
              osm_key: 'place',
              housenumber: '20',
              city: 'Meyzieu',
              street: 'Rue Edison',
              osm_value: 'house',
              postcode: '69330',
              type: 'house',
            },
          },
        ],
      };
      const result = await structuresService.searchAddress(data);

      expect(httpServiceMock.get).toHaveBeenCalledTimes(2);
      expect(result).toEqual(expectedResult);
    });
  });
});
