import { HttpService } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { Test, TestingModule } from '@nestjs/testing';
import { Types } from 'mongoose';
import { of } from 'rxjs';
import { PhotonResponseMock } from '../../test/mock/data/dataPhoton.mock.data';
import { structureDtoMock } from '../../test/mock/data/structure.mock.dto';
import { CategoriesServiceMock } from '../../test/mock/services/categories.mock.service';
import {
  StructuresExportServiceMock,
  mockFormattedStructures,
} from '../../test/mock/services/structures-export.mock.service';
import { StructuresServiceMock } from '../../test/mock/services/structures.mock.service';
import { TempUserServiceMock } from '../../test/mock/services/tempUser.mock.service';
import { UsersServiceMock } from '../../test/mock/services/user.mock.service';
import { CategoriesService } from '../categories/services/categories.service';
import { PersonalOffersService } from '../personal-offers/personal-offers.service';
import { TempUserService } from '../temp-user/temp-user.service';
import { UserRole } from '../users/enum/user-role.enum';
import { UsersService } from '../users/services/users.service';
import { PersonalOffersServiceMock } from './../../test/mock/services/personalOffers.mock.service';
import { PersonalOffer, PersonalOfferDocument } from './../personal-offers/schemas/personal-offer.schema';
import { CreateStructureDto } from './dto/create-structure.dto';
import { StructureFormattedInterface } from './interfaces/structure-formatted.interface';
import { StructuresExportService } from './services/structures-export.service';
import { StructuresService } from './services/structures.service';
import { StructuresController } from './structures.controller';

describe('StructuresController', () => {
  let structuresController: StructuresController;
  let usersService: UsersService;
  let structuresExportService: StructuresExportService;
  let personalOffersService: PersonalOffersService;

  const httpServiceMock = {
    get: jest.fn(),
  };

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [],
      controllers: [StructuresController],
      providers: [
        {
          provide: HttpService,
          useValue: httpServiceMock,
        },
        {
          provide: StructuresService,
          useClass: StructuresServiceMock,
        },
        {
          provide: UsersService,
          useClass: UsersServiceMock,
        },
        {
          provide: PersonalOffersService,
          useClass: PersonalOffersServiceMock,
        },
        {
          provide: TempUserService,
          useClass: TempUserServiceMock,
        },
        {
          provide: CategoriesService,
          useClass: CategoriesServiceMock,
        },
        {
          provide: StructuresExportService,
          useClass: StructuresExportServiceMock,
        },
      ],
    }).compile();
    structuresController = module.get<StructuresController>(StructuresController);
    usersService = module.get<UsersService>(UsersService);
    structuresExportService = module.get<StructuresExportService>(StructuresExportService);
    personalOffersService = module.get<PersonalOffersService>(PersonalOffersService);
  });

  it('should be defined', () => {
    expect(structuresController).toBeDefined();
  });

  it('should get structure coordinates', async () => {
    httpServiceMock.get.mockImplementationOnce(() => of(PhotonResponseMock));
    const coords = structuresController.getCoordinates('Lyon');
    expect(coords).toBeTruthy();
  });

  it('should create structure', async () => {
    const structure: CreateStructureDto = {
      userEmail: 'admin@admin.com',
      structure: structureDtoMock,
    };
    const res = structuresController.create(structure);
    expect(res).toBeTruthy();
  });

  it('should update structure after ownerVerify', async () => {
    const structureId = '1';
    const res = structuresController.updateAfterOwnerVerify(structureId);
    expect(res).toBeTruthy();
  });

  it('should update structure', async () => {
    // Useless test
    const structureService = new StructuresServiceMock();
    const structure = structureService.findOne(new Types.ObjectId('6093ba0e2ab5775cfc01ed3e'));

    const structureId = '1';
    const res = await structuresController.update(structureId, {
      deletedAt: null,
      remoteAccompaniment: null,
      dataShareConsentDate: null,
      personalOffers: [],
      ...structure,
    });
    expect(res.structureName).toBe('a');
  });

  it('should get all structure', async () => {
    const res = await structuresController.findAll();
    expect(res.length).toBe(2);
  });

  it('should get all Formatted structure', async () => {
    const mockResult: StructureFormattedInterface[] = mockFormattedStructures;
    jest.spyOn(structuresExportService, 'exportFormatted').mockResolvedValueOnce(mockResult);

    const result: StructureFormattedInterface[] = await structuresController.findAllFormatted();

    expect(result).toEqual(mockResult);
  });

  it('should claim structure', async () => {
    const userMock = new UsersServiceMock();
    const user = userMock.findOne('pauline.dupont@mii.com');
    const res = structuresController.claim('1', {
      phone: null,
      resetPasswordToken: null,
      changeEmailToken: null,
      newEmail: null,
      pendingStructuresLink: null,
      structuresLink: null,
      structureOutdatedMailSent: null,
      personalOffers: null,
      email: user.email,
      name: user.name,
      surname: user.surname,
      emailVerified: true,
      createdAt: new Date('2022-05-25T09:48:28.824Z'),
      password: user.password,
      validationToken: null,
      role: null,
      employer: {
        name: 'test',
        validated: true,
      },
      job: {
        name: 'test',
        validated: true,
        hasPersonalOffer: false,
        jobsGroup: null,
      },
      unattachedSince: null,
    });
    expect(res).toBeTruthy();
  });

  describe('find(structureId)', () => {
    it('should find deleted structure as an admin', async () => {
      const response = await structuresController.find({ user: { role: UserRole.admin } }, '6093ba0e2ab5775cfc01ed33');
      expect(response.deletedAt).toBeTruthy();
    });

    it('should not find deleted structure as anonymous', async () => {
      try {
        await structuresController.find({ user: {} }, '6093ba0e2ab5775cfc01ed33');
        expect(true).toBe(false);
      } catch (error) {
        expect(error.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });
  });

  describe('removeOwner', () => {
    it('should return invalid structure', async () => {
      try {
        await structuresController.removeOwner('invalidstructure', '63639058685ba134c32bc495');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Invalid Structure');
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });
    it('should return invalid user', async () => {
      try {
        await structuresController.removeOwner('6093ba0e2ab5775cfc01ed3e', 'invaliduser');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Invalid User');
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });
    it('should remove structure link without offer', async () => {
      jest.spyOn(usersService, 'getPersonalOfferInStructure').mockImplementationOnce(() => null);
      const removeFromStructureLinkedSpyer = jest.spyOn(usersService, 'removeFromStructureLinked');
      const deleteOfferSpyer = jest.spyOn(personalOffersService, 'delete');
      await structuresController.removeOwner('6093ba0e2ab5775cfc01ed3e', '63639058685ba134c32bc495');
      expect(removeFromStructureLinkedSpyer).toBeCalledTimes(1);
      expect(deleteOfferSpyer).toBeCalledTimes(0);
    });

    it('should remove structure link with offer', async () => {
      const result = ({
        _id: '2345ba0e2ab5775cfc01ed4d',
        categories: {
          onlineProcedures: ['caf', 'cpam'],
          baseSkills: [],
          advancedSkills: [],
        },
        createdAt: 'Wed Mar 16 2022 14:29:11 GMT+0100 (heure normale d’Europe centrale)',
        updatedAt: 'Wed Mar 16 2022 17:29:11 GMT+0100 (heure normale d’Europe centrale)',
      } as PersonalOffer) as PersonalOfferDocument;
      jest.spyOn(usersService, 'getPersonalOfferInStructure').mockImplementationOnce(() => result);
      const removeFromStructureLinkedSpyer = jest.spyOn(usersService, 'removeFromStructureLinked');
      const deleteOfferSpyer = jest.spyOn(personalOffersService, 'delete');
      await structuresController.removeOwner('6093ba0e2ab5775cfc01ed3e', '63639058685ba134c32bc495');
      expect(removeFromStructureLinkedSpyer).toBeCalledTimes(1);
      expect(deleteOfferSpyer).toBeCalledTimes(1);
    });
  });
});
